use Test::More tests=>3;

use strict;
use warnings;

use_ok 'MooseX::Types::Parameterizable';
use_ok 'MooseX::Meta::TypeConstraint::Parameterizable';
use_ok 'MooseX::Meta::TypeCoercion::Parameterizable';

